@@ -31,41 +31,26 @@ static void print_dir_item_type(struct extent_buffer *eb,
                                 struct btrfs_dir_item *di)
 {
 	u8 type = btrfs_dir_type(eb, di);
+	static const char* dir_item_str[] = {
+		[BTRFS_FT_REG_FILE]	= "FILE",
+		[BTRFS_FT_DIR] 		= "DIR",
+		[BTRFS_FT_CHRDEV]	= "CHRDEV",
+		[BTRFS_FT_BLKDEV]	= "BLKDEV",
+		[BTRFS_FT_FIFO]		= "FIFO",
+		[BTRFS_FT_SOCK]		= "SOCK",
+		[BTRFS_FT_SYMLINK]	= "SYMLINK",
+		[BTRFS_FT_XATTR]	= "XATTR"
+	};
 
-	switch (type) {
-	case BTRFS_FT_REG_FILE:
-		printf("FILE");
-		break;
-	case BTRFS_FT_DIR:
-		printf("DIR");
-		break;
-	case BTRFS_FT_CHRDEV:
-		printf("CHRDEV");
-		break;
-	case BTRFS_FT_BLKDEV:
-		printf("BLKDEV");
-		break;
-	case BTRFS_FT_FIFO:
-		printf("FIFO");
-		break;
-	case BTRFS_FT_SOCK:
-		printf("SOCK");
-		break;
-	case BTRFS_FT_SYMLINK:
-		printf("SYMLINK");
-		break;
-	case BTRFS_FT_XATTR:
-		printf("XATTR");
-		break;
-	default:
-		printf("%u", type);
-	}
+	if (type < ARRAY_SIZE(dir_item_str) && dir_item_str[type])
+		printf("%s", dir_item_str[type]);
+	else
+		printf("DIR_ITEM.%u", type);
 }
 
-static int print_dir_item(struct extent_buffer *eb, struct btrfs_item *item,
+static void print_dir_item(struct extent_buffer *eb, u32 size,
 			  struct btrfs_dir_item *di)
 {
-	u32 total;
 	u32 cur = 0;
 	u32 len;
 	u32 name_len;
@@ -73,8 +58,7 @@ static int print_dir_item(struct extent_buffer *eb, struct btrfs_item *item,
 	char namebuf[BTRFS_NAME_LEN];
 	struct btrfs_disk_key location;
 
-	total = btrfs_item_size(eb, item);
-	while(cur < total) {
+	while (cur < size) {
 		btrfs_dir_item_key(eb, di, &location);
 		printf("\t\tlocation ");
 		btrfs_print_key(&location);
@@ -85,8 +69,10 @@ static int print_dir_item(struct extent_buffer *eb, struct btrfs_item *item,
 		data_len = btrfs_dir_data_len(eb, di);
 		len = (name_len <= sizeof(namebuf))? name_len: sizeof(namebuf);
 		read_extent_buffer(eb, namebuf, (unsigned long)(di + 1), len);
-		printf("\t\tnamelen %u datalen %u name: %.*s\n",
-		       name_len, data_len, len, namebuf);
+		printf("\t\ttransid %llu data_len %u name_len %u\n",
+				btrfs_dir_transid(eb, di),
+				data_len, name_len);
+		printf("\t\tname: %.*s\n", len, namebuf);
 		if (data_len) {
 			len = (data_len <= sizeof(namebuf))? data_len: sizeof(namebuf);
 			read_extent_buffer(eb, namebuf,
@@ -97,14 +83,11 @@ static int print_dir_item(struct extent_buffer *eb, struct btrfs_item *item,
 		di = (struct btrfs_dir_item *)((char *)di + len);
 		cur += len;
 	}
-	return 0;
 }
 
-static int print_inode_extref_item(struct extent_buffer *eb,
-				   struct btrfs_item *item,
-				   struct btrfs_inode_extref *extref)
+static void print_inode_extref_item(struct extent_buffer *eb, u32 size,
+		struct btrfs_inode_extref *extref)
 {
-	u32 total;
 	u32 cur = 0;
 	u32 len;
 	u32 name_len = 0;
@@ -112,9 +95,7 @@ static int print_inode_extref_item(struct extent_buffer *eb,
 	u64 parent_objid;
 	char namebuf[BTRFS_NAME_LEN];
 
-	total = btrfs_item_size(eb, item);
-
-	while (cur < total) {
+	while (cur < size) {
 		index = btrfs_inode_extref_index(eb, extref);
 		name_len = btrfs_inode_extref_name_len(eb, extref);
 		parent_objid = btrfs_inode_extref_parent(eb, extref);
@@ -133,20 +114,18 @@ static int print_inode_extref_item(struct extent_buffer *eb,
 		extref = (struct btrfs_inode_extref *)((char *)extref + len);
 		cur += len;
 	}
-	return 0;
 }
 
-static int print_inode_ref_item(struct extent_buffer *eb, struct btrfs_item *item,
+static void print_inode_ref_item(struct extent_buffer *eb, u32 size,
 				struct btrfs_inode_ref *ref)
 {
-	u32 total;
 	u32 cur = 0;
 	u32 len;
 	u32 name_len;
 	u64 index;
 	char namebuf[BTRFS_NAME_LEN];
-	total = btrfs_item_size(eb, item);
-	while(cur < total) {
+
+	while (cur < size) {
 		name_len = btrfs_inode_ref_name_len(eb, ref);
 		index = btrfs_inode_ref_index(eb, ref);
 		len = (name_len <= sizeof(namebuf))? name_len: sizeof(namebuf);
@@ -157,7 +136,6 @@ static int print_inode_ref_item(struct extent_buffer *eb, struct btrfs_item *ite
 		ref = (struct btrfs_inode_ref *)((char *)ref + len);
 		cur += len;
 	}
-	return 0;
 }
 
 /* Caller should ensure sizeof(*ret)>=21 "DATA|METADATA|RAID10" */
@@ -224,12 +202,17 @@ void print_chunk(struct extent_buffer *eb, struct btrfs_chunk *chunk)
 	char chunk_flags_str[32] = {0};
 
 	bg_flags_to_str(btrfs_chunk_type(eb, chunk), chunk_flags_str);
-	printf("\t\tchunk length %llu owner %llu stripe_len %llu\n",
+	printf("\t\tlength %llu owner %llu stripe_len %llu type %s\n",
 	       (unsigned long long)btrfs_chunk_length(eb, chunk),
 	       (unsigned long long)btrfs_chunk_owner(eb, chunk),
-	       (unsigned long long)btrfs_chunk_stripe_len(eb, chunk));
-	printf("\t\ttype %s num_stripes %d\n",
-	       chunk_flags_str, num_stripes);
+	       (unsigned long long)btrfs_chunk_stripe_len(eb, chunk),
+		chunk_flags_str);
+	printf("\t\tio_align %u io_width %u sector_size %u\n",
+			btrfs_chunk_io_align(eb, chunk),
+			btrfs_chunk_io_width(eb, chunk),
+			btrfs_chunk_sector_size(eb, chunk));
+	printf("\t\tnum_stripes %hu sub_stripes %hu\n", num_stripes,
+			btrfs_chunk_sub_stripes(eb, chunk));
 	for (i = 0 ; i < num_stripes ; i++) {
 		unsigned char dev_uuid[BTRFS_UUID_SIZE];
 		char str_dev_uuid[BTRFS_UUID_UNPARSED_SIZE];
@@ -241,27 +224,45 @@ void print_chunk(struct extent_buffer *eb, struct btrfs_chunk *chunk)
 		printf("\t\t\tstripe %d devid %llu offset %llu\n", i,
 		      (unsigned long long)btrfs_stripe_devid_nr(eb, chunk, i),
 		      (unsigned long long)btrfs_stripe_offset_nr(eb, chunk, i));
-		printf("\t\t\tdev uuid: %s\n", str_dev_uuid);
+		printf("\t\t\tdev_uuid %s\n", str_dev_uuid);
 	}
 }
 
 static void print_dev_item(struct extent_buffer *eb,
 			   struct btrfs_dev_item *dev_item)
 {
-	char disk_uuid_c[BTRFS_UUID_UNPARSED_SIZE];
-	u8 disk_uuid[BTRFS_UUID_SIZE];
+	char uuid_str[BTRFS_UUID_UNPARSED_SIZE];
+	char fsid_str[BTRFS_UUID_UNPARSED_SIZE];
+	u8 uuid[BTRFS_UUID_SIZE];
+	u8 fsid[BTRFS_UUID_SIZE];
 
-	read_extent_buffer(eb, disk_uuid,
+	read_extent_buffer(eb, uuid,
 			   (unsigned long)btrfs_device_uuid(dev_item),
 			   BTRFS_UUID_SIZE);
-	uuid_unparse(disk_uuid, disk_uuid_c);
-	printf("\t\tdev item devid %llu "
-	       "total_bytes %llu bytes used %Lu\n"
-	       "\t\tdev uuid %s\n",
+	uuid_unparse(uuid, uuid_str);
+	read_extent_buffer(eb, fsid,
+			   (unsigned long)btrfs_device_fsid(dev_item),
+			   BTRFS_UUID_SIZE);
+	uuid_unparse(fsid, fsid_str);
+	printf("\t\tdevid %llu total_bytes %llu bytes_used %Lu\n"
+	       "\t\tio_align %u io_width %u sector_size %u type %llu\n"
+	       "\t\tgeneration %llu start_offset %llu dev_group %u\n"
+	       "\t\tseek_speed %hhu bandwidth %hhu\n"
+	       "\t\tuuid %s\n"
+	       "\t\tfsid %s\n",
 	       (unsigned long long)btrfs_device_id(eb, dev_item),
 	       (unsigned long long)btrfs_device_total_bytes(eb, dev_item),
 	       (unsigned long long)btrfs_device_bytes_used(eb, dev_item),
-	       disk_uuid_c);
+	       btrfs_device_io_align(eb, dev_item),
+	       btrfs_device_io_width(eb, dev_item),
+	       btrfs_device_sector_size(eb, dev_item),
+	       (unsigned long long)btrfs_device_type(eb, dev_item),
+	       (unsigned long long)btrfs_device_generation(eb, dev_item),
+	       (unsigned long long)btrfs_device_start_offset(eb, dev_item),
+	       btrfs_device_group(eb, dev_item),
+	       btrfs_device_seek_speed(eb, dev_item),
+	       btrfs_device_bandwidth(eb, dev_item),
+	       uuid_str, fsid_str);
 }
 
 static void print_uuids(struct extent_buffer *eb)
@@ -302,6 +303,16 @@ static void compress_type_to_str(u8 compress_type, char *ret)
 	}
 }
 
+static const char* file_extent_type_to_str(u8 type)
+{
+	switch (type) {
+	case BTRFS_FILE_EXTENT_INLINE: return "inline";
+	case BTRFS_FILE_EXTENT_PREALLOC: return "prealloc";
+	case BTRFS_FILE_EXTENT_REG: return "regular";
+	default: return "unknown";
+	}
+}
+
 static void print_file_extent_item(struct extent_buffer *eb,
 				   struct btrfs_item *item,
 				   int slot,
@@ -313,12 +324,16 @@ static void print_file_extent_item(struct extent_buffer *eb,
 	compress_type_to_str(btrfs_file_extent_compression(eb, fi),
 			     compress_str);
 
+	printf("\t\tgeneration %llu type %hhu (%s)\n",
+			btrfs_file_extent_generation(eb, fi),
+			extent_type, file_extent_type_to_str(extent_type));
+
 	if (extent_type == BTRFS_FILE_EXTENT_INLINE) {
-		printf("\t\tinline extent data size %u "
-		       "ram %u compress(%s)\n",
-		  btrfs_file_extent_inline_item_len(eb, item),
-		  btrfs_file_extent_inline_len(eb, slot, fi),
-		  compress_str);
+		printf("\t\tinline extent data size %u ram_bytes %u compression %hhu (%s)\n",
+				btrfs_file_extent_inline_item_len(eb, item),
+				btrfs_file_extent_inline_len(eb, slot, fi),
+				btrfs_file_extent_compression(eb, fi),
+				compress_str);
 		return;
 	}
 	if (extent_type == BTRFS_FILE_EXTENT_PREALLOC) {
@@ -337,7 +352,9 @@ static void print_file_extent_item(struct extent_buffer *eb,
 		(unsigned long long)btrfs_file_extent_offset(eb, fi),
 		(unsigned long long)btrfs_file_extent_num_bytes(eb, fi),
 		(unsigned long long)btrfs_file_extent_ram_bytes(eb, fi));
-	printf("\t\textent compression(%s)\n", compress_str);
+	printf("\t\textent compression %hhu (%s)\n",
+			btrfs_file_extent_compression(eb, fi),
+			compress_str);
 }
 
 /* Caller should ensure sizeof(*ret) >= 16("DATA|TREE_BLOCK") */
@@ -470,7 +487,7 @@ static void print_extent_ref_v0(struct extent_buffer *eb, int slot)
 }
 #endif
 
-static void print_root_ref(struct extent_buffer *leaf, int slot, char *tag)
+static void print_root_ref(struct extent_buffer *leaf, int slot, const char *tag)
 {
 	struct btrfs_root_ref *ref;
 	char namebuf[BTRFS_NAME_LEN];
@@ -485,15 +502,14 @@ static void print_root_ref(struct extent_buffer *leaf, int slot, char *tag)
 	       namelen, namebuf);
 }
 
-static int count_bytes(void *buf, int len, char b)
+static int empty_uuid(const u8 *uuid)
 {
-	int cnt = 0;
 	int i;
-	for (i = 0; i < len; i++) {
-		if (((char*)buf)[i] == b)
-			cnt++;
-	}
-	return cnt;
+
+	for (i = 0; i < BTRFS_UUID_SIZE; i++)
+		if (uuid[i])
+			return 0;
+	return 1;
 }
 
 /*
@@ -514,6 +530,7 @@ static void print_root(struct extent_buffer *leaf, int slot)
 	int len;
 	char uuid_str[BTRFS_UUID_UNPARSED_SIZE];
 	char flags_str[32] = {0};
+	struct btrfs_key drop_key;
 
 	ri = btrfs_item_ptr(leaf, slot, struct btrfs_root_item);
 	len = btrfs_item_size_nr(leaf, slot);
@@ -522,24 +539,27 @@ static void print_root(struct extent_buffer *leaf, int slot)
 	read_extent_buffer(leaf, &root_item, (unsigned long)ri, len);
 	root_flags_to_str(btrfs_root_flags(&root_item), flags_str);
 
-	printf("\t\troot data bytenr %llu level %d dirid %llu refs %u gen %llu lastsnap %llu\n",
+	printf("\t\tgeneration %llu root_dirid %llu bytenr %llu level %hhu refs %u\n",
+		(unsigned long long)btrfs_root_generation(&root_item),
+		(unsigned long long)btrfs_root_dirid(&root_item),
 		(unsigned long long)btrfs_root_bytenr(&root_item),
 		btrfs_root_level(&root_item),
-		(unsigned long long)btrfs_root_dirid(&root_item),
-		btrfs_root_refs(&root_item),
-		(unsigned long long)btrfs_root_generation(&root_item),
-		(unsigned long long)btrfs_root_last_snapshot(&root_item));
-	printf("\t\tflags 0x%llx(%s)\n", btrfs_root_flags(&root_item),
-	       flags_str);
+		btrfs_root_refs(&root_item));
+	printf("\t\tlastsnap %llu byte_limit %llu bytes_used %llu flags 0x%llx(%s)\n",
+		(unsigned long long)btrfs_root_last_snapshot(&root_item),
+		(unsigned long long)btrfs_root_limit(&root_item),
+		(unsigned long long)btrfs_root_used(&root_item),
+		(unsigned long long)btrfs_root_flags(&root_item),
+		flags_str);
 
 	if (root_item.generation == root_item.generation_v2) {
 		uuid_unparse(root_item.uuid, uuid_str);
 		printf("\t\tuuid %s\n", uuid_str);
-		if (count_bytes(root_item.parent_uuid, BTRFS_UUID_SIZE, 0) != BTRFS_UUID_SIZE) {
+		if (!empty_uuid(root_item.parent_uuid)) {
 			uuid_unparse(root_item.parent_uuid, uuid_str);
 			printf("\t\tparent_uuid %s\n", uuid_str);
 		}
-		if (count_bytes(root_item.received_uuid, BTRFS_UUID_SIZE, 0) != BTRFS_UUID_SIZE) {
+		if (!empty_uuid(root_item.received_uuid)) {
 			uuid_unparse(root_item.received_uuid, uuid_str);
 			printf("\t\treceived_uuid %s\n", uuid_str);
 		}
@@ -551,14 +571,11 @@ static void print_root(struct extent_buffer *leaf, int slot)
 				btrfs_root_rtransid(&root_item));
 		}
 	}
-	if (btrfs_root_refs(&root_item) == 0) {
-		struct btrfs_key drop_key;
-		btrfs_disk_key_to_cpu(&drop_key,
-				      &root_item.drop_progress);
-		printf("\t\tdrop ");
-		btrfs_print_key(&root_item.drop_progress);
-		printf(" level %d\n", root_item.drop_level);
-	}
+
+	btrfs_disk_key_to_cpu(&drop_key, &root_item.drop_progress);
+	printf("\t\tdrop ");
+	btrfs_print_key(&root_item.drop_progress);
+	printf(" level %hhu\n", root_item.drop_level);
 }
 
 static void print_free_space_header(struct extent_buffer *leaf, int slot)
@@ -579,132 +596,58 @@ static void print_free_space_header(struct extent_buffer *leaf, int slot)
 
 void print_key_type(FILE *stream, u64 objectid, u8 type)
 {
+	static const char* key_to_str[256] = {
+		[BTRFS_INODE_ITEM_KEY]		= "INODE_ITEM",
+		[BTRFS_INODE_REF_KEY]		= "INODE_REF",
+		[BTRFS_INODE_EXTREF_KEY]	= "INODE_EXTREF",
+		[BTRFS_DIR_ITEM_KEY]		= "DIR_ITEM",
+		[BTRFS_DIR_INDEX_KEY]		= "DIR_INDEX",
+		[BTRFS_DIR_LOG_ITEM_KEY]	= "DIR_LOG_ITEM",
+		[BTRFS_DIR_LOG_INDEX_KEY]	= "DIR_LOG_INDEX",
+		[BTRFS_XATTR_ITEM_KEY]		= "XATTR_ITEM",
+		[BTRFS_ORPHAN_ITEM_KEY]		= "ORPHAN_ITEM",
+		[BTRFS_ROOT_ITEM_KEY]		= "ROOT_ITEM",
+		[BTRFS_ROOT_REF_KEY]		= "ROOT_REF",
+		[BTRFS_ROOT_BACKREF_KEY]	= "ROOT_BACKREF",
+		[BTRFS_EXTENT_ITEM_KEY]		= "EXTENT_ITEM",
+		[BTRFS_METADATA_ITEM_KEY]	= "METADATA_ITEM",
+		[BTRFS_TREE_BLOCK_REF_KEY]	= "TREE_BLOCK_REF",
+		[BTRFS_SHARED_BLOCK_REF_KEY]	= "SHARED_BLOCK_REF",
+		[BTRFS_EXTENT_DATA_REF_KEY]	= "EXTENT_DATA_REF",
+		[BTRFS_SHARED_DATA_REF_KEY]	= "SHARED_DATA_REF",
+		[BTRFS_EXTENT_REF_V0_KEY]	= "EXTENT_REF_V0",
+		[BTRFS_CSUM_ITEM_KEY]		= "CSUM_ITEM",
+		[BTRFS_EXTENT_CSUM_KEY]		= "EXTENT_CSUM",
+		[BTRFS_EXTENT_DATA_KEY]		= "EXTENT_DATA",
+		[BTRFS_BLOCK_GROUP_ITEM_KEY]	= "BLOCK_GROUP_ITEM",
+		[BTRFS_FREE_SPACE_INFO_KEY]	= "FREE_SPACE_INFO",
+		[BTRFS_FREE_SPACE_EXTENT_KEY]	= "FREE_SPACE_EXTENT",
+		[BTRFS_FREE_SPACE_BITMAP_KEY]	= "FREE_SPACE_BITMAP",
+		[BTRFS_CHUNK_ITEM_KEY]		= "CHUNK_ITEM",
+		[BTRFS_DEV_ITEM_KEY]		= "DEV_ITEM",
+		[BTRFS_DEV_EXTENT_KEY]		= "DEV_EXTENT",
+		[BTRFS_TEMPORARY_ITEM_KEY]	= "TEMPORARY_ITEM",
+		[BTRFS_DEV_REPLACE_KEY]		= "DEV_REPLACE",
+		[BTRFS_STRING_ITEM_KEY]		= "STRING_ITEM",
+		[BTRFS_QGROUP_STATUS_KEY]	= "QGROUP_STATUS",
+		[BTRFS_QGROUP_RELATION_KEY]	= "QGROUP_RELATION",
+		[BTRFS_QGROUP_INFO_KEY]		= "QGROUP_INFO",
+		[BTRFS_QGROUP_LIMIT_KEY]	= "QGROUP_LIMIT",
+		[BTRFS_PERSISTENT_ITEM_KEY]	= "PERSISTENT_ITEM",
+		[BTRFS_UUID_KEY_SUBVOL]		= "UUID_KEY_SUBVOL",
+		[BTRFS_UUID_KEY_RECEIVED_SUBVOL] = "UUID_KEY_RECEIVED_SUBVOL",
+	};
+
 	if (type == 0 && objectid == BTRFS_FREE_SPACE_OBJECTID) {
 		fprintf(stream, "UNTYPED");
 		return;
 	}
 
-	switch (type) {
-	case BTRFS_INODE_ITEM_KEY:
-		fprintf(stream, "INODE_ITEM");
-		break;
-	case BTRFS_INODE_REF_KEY:
-		fprintf(stream, "INODE_REF");
-		break;
-	case BTRFS_INODE_EXTREF_KEY:
-		fprintf(stream, "INODE_EXTREF");
-		break;
-	case BTRFS_DIR_ITEM_KEY:
-		fprintf(stream, "DIR_ITEM");
-		break;
-	case BTRFS_DIR_INDEX_KEY:
-		fprintf(stream, "DIR_INDEX");
-		break;
-	case BTRFS_DIR_LOG_ITEM_KEY:
-		fprintf(stream, "DIR_LOG_ITEM");
-		break;
-	case BTRFS_DIR_LOG_INDEX_KEY:
-		fprintf(stream, "DIR_LOG_INDEX");
-		break;
-	case BTRFS_XATTR_ITEM_KEY:
-		fprintf(stream, "XATTR_ITEM");
-		break;
-	case BTRFS_ORPHAN_ITEM_KEY:
-		fprintf(stream, "ORPHAN_ITEM");
-		break;
-	case BTRFS_ROOT_ITEM_KEY:
-		fprintf(stream, "ROOT_ITEM");
-		break;
-	case BTRFS_ROOT_REF_KEY:
-		fprintf(stream, "ROOT_REF");
-		break;
-	case BTRFS_ROOT_BACKREF_KEY:
-		fprintf(stream, "ROOT_BACKREF");
-		break;
-	case BTRFS_EXTENT_ITEM_KEY:
-		fprintf(stream, "EXTENT_ITEM");
-		break;
-	case BTRFS_METADATA_ITEM_KEY:
-		fprintf(stream, "METADATA_ITEM");
-		break;
-	case BTRFS_TREE_BLOCK_REF_KEY:
-		fprintf(stream, "TREE_BLOCK_REF");
-		break;
-	case BTRFS_SHARED_BLOCK_REF_KEY:
-		fprintf(stream, "SHARED_BLOCK_REF");
-		break;
-	case BTRFS_EXTENT_DATA_REF_KEY:
-		fprintf(stream, "EXTENT_DATA_REF");
-		break;
-	case BTRFS_SHARED_DATA_REF_KEY:
-		fprintf(stream, "SHARED_DATA_REF");
-		break;
-	case BTRFS_EXTENT_REF_V0_KEY:
-		fprintf(stream, "EXTENT_REF_V0");
-		break;
-	case BTRFS_CSUM_ITEM_KEY:
-		fprintf(stream, "CSUM_ITEM");
-		break;
-	case BTRFS_EXTENT_CSUM_KEY:
-		fprintf(stream, "EXTENT_CSUM");
-		break;
-	case BTRFS_EXTENT_DATA_KEY:
-		fprintf(stream, "EXTENT_DATA");
-		break;
-	case BTRFS_BLOCK_GROUP_ITEM_KEY:
-		fprintf(stream, "BLOCK_GROUP_ITEM");
-		break;
-	case BTRFS_FREE_SPACE_INFO_KEY:
-		fprintf(stream, "FREE_SPACE_INFO");
-		break;
-	case BTRFS_FREE_SPACE_EXTENT_KEY:
-		fprintf(stream, "FREE_SPACE_EXTENT");
-		break;
-	case BTRFS_FREE_SPACE_BITMAP_KEY:
-		fprintf(stream, "FREE_SPACE_BITMAP");
-		break;
-	case BTRFS_CHUNK_ITEM_KEY:
-		fprintf(stream, "CHUNK_ITEM");
-		break;
-	case BTRFS_DEV_ITEM_KEY:
-		fprintf(stream, "DEV_ITEM");
-		break;
-	case BTRFS_DEV_EXTENT_KEY:
-		fprintf(stream, "DEV_EXTENT");
-		break;
-	case BTRFS_BALANCE_ITEM_KEY:
-		fprintf(stream, "BALANCE_ITEM");
-		break;
-	case BTRFS_DEV_REPLACE_KEY:
-		fprintf(stream, "DEV_REPLACE");
-		break;
-	case BTRFS_STRING_ITEM_KEY:
-		fprintf(stream, "STRING_ITEM");
-		break;
-	case BTRFS_QGROUP_STATUS_KEY:
-		fprintf(stream, "QGROUP_STATUS");
-		break;
-	case BTRFS_QGROUP_RELATION_KEY:
-		fprintf(stream, "QGROUP_RELATION");
-		break;
-	case BTRFS_QGROUP_INFO_KEY:
-		fprintf(stream, "QGROUP_INFO");
-		break;
-	case BTRFS_QGROUP_LIMIT_KEY:
-		fprintf(stream, "QGROUP_LIMIT");
-		break;
-	case BTRFS_DEV_STATS_KEY:
-		fprintf(stream, "DEV_STATS");
-		break;
-	case BTRFS_UUID_KEY_SUBVOL:
-		fprintf(stream, "UUID_KEY_SUBVOL");
-		break;
-	case BTRFS_UUID_KEY_RECEIVED_SUBVOL:
-		fprintf(stream, "UUID_KEY_RECEIVED_SUBVOL");
-		break;
-	default:
+
+	if (key_to_str[type])
+		fputs(key_to_str[type], stream);
+	else
 		fprintf(stream, "UNKNOWN.%d", type);
-	};
 }
 
 void print_objectid(FILE *stream, u64 objectid, u8 type)
@@ -851,138 +794,237 @@ static void print_uuid_item(struct extent_buffer *l, unsigned long offset,
 	}
 }
 
-/* Caller should ensure sizeof(*ret) >= 29 "NODATASUM|NODATACOW|READONLY" */
+/* Btrfs inode flag stringification helper */
+#define STRCAT_ONE_INODE_FLAG(flags, name, empty, dst) ({			\
+	if (flags & BTRFS_INODE_##name) {				\
+		if (!empty)						\
+			strcat(dst, "|");				\
+		strcat(dst, #name);					\
+		empty = 0;						\
+	}								\
+})
+
+/*
+ * Caller should ensure sizeof(*ret) >= 102: all charactors plus '|' of
+ * BTRFS_INODE_* flags
+ */
 static void inode_flags_to_str(u64 flags, char *ret)
 {
 	int empty = 1;
 
-	if (flags & BTRFS_INODE_NODATASUM) {
-		empty = 0;
-		strcpy(ret, "NODATASUM");
-	}
-	if (flags & BTRFS_INODE_NODATACOW) {
-		if (!empty) {
-			empty = 0;
-			strcat(ret, "|");
-		}
-		strcat(ret, "NODATACOW");
-	}
-	if (flags & BTRFS_INODE_READONLY) {
-		if (!empty) {
-			empty = 0;
-			strcat(ret, "|");
-		}
-		strcat(ret, "READONLY");
-	}
+	STRCAT_ONE_INODE_FLAG(flags, NODATASUM, empty, ret);
+	STRCAT_ONE_INODE_FLAG(flags, NODATACOW, empty, ret);
+	STRCAT_ONE_INODE_FLAG(flags, READONLY, empty, ret);
+	STRCAT_ONE_INODE_FLAG(flags, NOCOMPRESS, empty, ret);
+	STRCAT_ONE_INODE_FLAG(flags, PREALLOC, empty, ret);
+	STRCAT_ONE_INODE_FLAG(flags, SYNC, empty, ret);
+	STRCAT_ONE_INODE_FLAG(flags, IMMUTABLE, empty, ret);
+	STRCAT_ONE_INODE_FLAG(flags, APPEND, empty, ret);
+	STRCAT_ONE_INODE_FLAG(flags, NODUMP, empty, ret);
+	STRCAT_ONE_INODE_FLAG(flags, NOATIME, empty, ret);
+	STRCAT_ONE_INODE_FLAG(flags, DIRSYNC, empty, ret);
+	STRCAT_ONE_INODE_FLAG(flags, COMPRESS, empty, ret);
 	if (empty)
 		strcat(ret, "none");
 }
 
-void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
+static void print_timespec(struct extent_buffer *eb,
+		struct btrfs_timespec *timespec, const char *prefix,
+		const char *suffix)
+{
+	struct tm tm;
+	u64 tmp_u64;
+	u32 tmp_u32;
+	time_t tmp_time;
+	char timestamp[256];
+
+	tmp_u64 = btrfs_timespec_sec(eb, timespec);
+	tmp_u32 = btrfs_timespec_nsec(eb, timespec);
+	tmp_time = tmp_u64;
+	localtime_r(&tmp_time, &tm);
+	strftime(timestamp, sizeof(timestamp),
+			"%Y-%m-%d %H:%M:%S", &tm);
+	printf("%s%llu.%u (%s)%s", prefix, (unsigned long long)tmp_u64, tmp_u32,
+			timestamp, suffix);
+}
+
+static void print_inode_item(struct extent_buffer *eb,
+		struct btrfs_inode_item *ii)
+{
+	char flags_str[256];
+
+	memset(flags_str, 0, sizeof(flags_str));
+	inode_flags_to_str(btrfs_inode_flags(eb, ii), flags_str);
+	printf("\t\tinode generation %llu transid %llu size %llu nbytes %llu\n"
+	       "\t\tblock group %llu mode %o links %u uid %u gid %u rdev %llu\n"
+	       "\t\tsequence %llu flags 0x%llx(%s)\n",
+	       (unsigned long long)btrfs_inode_generation(eb, ii),
+	       (unsigned long long)btrfs_inode_transid(eb, ii),
+	       (unsigned long long)btrfs_inode_size(eb, ii),
+	       (unsigned long long)btrfs_inode_nbytes(eb, ii),
+	       (unsigned long long)btrfs_inode_block_group(eb,ii),
+	       btrfs_inode_mode(eb, ii),
+	       btrfs_inode_nlink(eb, ii),
+	       btrfs_inode_uid(eb, ii),
+	       btrfs_inode_gid(eb, ii),
+	       (unsigned long long)btrfs_inode_rdev(eb,ii),
+	       (unsigned long long)btrfs_inode_flags(eb,ii),
+	       (unsigned long long)btrfs_inode_sequence(eb, ii),
+	       flags_str);
+	print_timespec(eb, btrfs_inode_atime(ii), "\t\tatime ", "\n");
+	print_timespec(eb, btrfs_inode_ctime(ii), "\t\tctime ", "\n");
+	print_timespec(eb, btrfs_inode_mtime(ii), "\t\tmtime ", "\n");
+	print_timespec(eb, btrfs_inode_otime(ii), "\t\totime ", "\n");
+}
+
+static void print_disk_balance_args(struct btrfs_disk_balance_args *ba)
+{
+	printf("\t\tprofiles %llu devid %llu target %llu flags %llu\n",
+			(unsigned long long)le64_to_cpu(ba->profiles),
+			(unsigned long long)le64_to_cpu(ba->devid),
+			(unsigned long long)le64_to_cpu(ba->target),
+			(unsigned long long)le64_to_cpu(ba->flags));
+	printf("\t\tusage_min %u usage_max %u pstart %llu pend %llu\n",
+			le32_to_cpu(ba->usage_min),
+			le32_to_cpu(ba->usage_max),
+			(unsigned long long)le64_to_cpu(ba->pstart),
+			(unsigned long long)le64_to_cpu(ba->pend));
+	printf("\t\tvstart %llu vend %llu limit_min %u limit_max %u\n",
+			(unsigned long long)le64_to_cpu(ba->vstart),
+			(unsigned long long)le64_to_cpu(ba->vend),
+			le32_to_cpu(ba->limit_min),
+			le32_to_cpu(ba->limit_max));
+	printf("\t\tstripes_min %u stripes_max %u\n",
+			le32_to_cpu(ba->stripes_min),
+			le32_to_cpu(ba->stripes_max));
+}
+
+static void print_balance_item(struct extent_buffer *eb,
+		struct btrfs_balance_item *bi)
+{
+	printf("\t\tbalance status flags %llu\n",
+			btrfs_balance_item_flags(eb, bi));
+
+	printf("\t\tDATA\n");
+	print_disk_balance_args(btrfs_balance_item_data(eb, bi));
+	printf("\t\tMETADATA\n");
+	print_disk_balance_args(btrfs_balance_item_meta(eb, bi));
+	printf("\t\tSYSTEM\n");
+	print_disk_balance_args(btrfs_balance_item_sys(eb, bi));
+}
+
+static void print_dev_stats(struct extent_buffer *eb,
+		struct btrfs_dev_stats_item *stats, u32 size)
 {
 	int i;
-	char *str;
+	u32 known = BTRFS_DEV_STAT_VALUES_MAX * sizeof(__le64);
+	__le64 *values = btrfs_dev_stats_values(eb, stats);
+
+	printf("\t\tdevice stats\n");
+	printf("\t\twrite_errs %llu read_errs %llu flush_errs %llu corruption_errs %llu generation %llu\n",
+		(unsigned long long)le64_to_cpu(values[BTRFS_DEV_STAT_WRITE_ERRS]),
+		(unsigned long long)le64_to_cpu(values[BTRFS_DEV_STAT_READ_ERRS]),
+		(unsigned long long)le64_to_cpu(values[BTRFS_DEV_STAT_FLUSH_ERRS]),
+		(unsigned long long)le64_to_cpu(values[BTRFS_DEV_STAT_CORRUPTION_ERRS]),
+		(unsigned long long)le64_to_cpu(values[BTRFS_DEV_STAT_GENERATION_ERRS]));
+
+	if (known < size) {
+		printf("\t\tunknown stats item bytes %u", size - known);
+		for (i = BTRFS_DEV_STAT_VALUES_MAX; i * sizeof(__le64) < size; i++) {
+			printf("\t\tunknown item %u offset %lu value %llu\n",
+				i, i * sizeof(__le64),
+				(unsigned long long)le64_to_cpu(values[i]));
+		}
+	}
+}
+
+void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *eb)
+{
 	struct btrfs_item *item;
-	struct btrfs_dir_item *di;
-	struct btrfs_inode_item *ii;
-	struct btrfs_file_extent_item *fi;
-	struct btrfs_block_group_item *bi;
-	struct btrfs_extent_data_ref *dref;
-	struct btrfs_shared_data_ref *sref;
-	struct btrfs_inode_ref *iref;
-	struct btrfs_inode_extref *iref2;
-	struct btrfs_dev_extent *dev_extent;
 	struct btrfs_disk_key disk_key;
-	struct btrfs_block_group_item bg_item;
-	struct btrfs_free_space_info *free_info;
-	struct btrfs_dir_log_item *dlog;
-	struct btrfs_qgroup_info_item *qg_info;
-	struct btrfs_qgroup_limit_item *qg_limit;
-	struct btrfs_qgroup_status_item *qg_status;
-	u32 nr = btrfs_header_nritems(l);
-	u64 objectid;
-	u32 type;
-	char flags_str[32];
+	u32 i;
+	u32 nr;
+
+	nr = btrfs_header_nritems(eb);
 
 	printf("leaf %llu items %d free space %d generation %llu owner %llu\n",
-		(unsigned long long)btrfs_header_bytenr(l), nr,
-		btrfs_leaf_free_space(root, l),
-		(unsigned long long)btrfs_header_generation(l),
-		(unsigned long long)btrfs_header_owner(l));
-	print_uuids(l);
+		(unsigned long long)btrfs_header_bytenr(eb), nr,
+		btrfs_leaf_free_space(root, eb),
+		(unsigned long long)btrfs_header_generation(eb),
+		(unsigned long long)btrfs_header_owner(eb));
+	print_uuids(eb);
 	fflush(stdout);
-	for (i = 0 ; i < nr ; i++) {
+
+	for (i = 0; i < nr; i++) {
+		u32 item_size;
+		void *ptr;
+		u64 objectid;
+		u32 type;
+		u64 offset;
+		char flags_str[256];
+		char uuid_str[BTRFS_UUID_UNPARSED_SIZE];
+		u8 uuid[BTRFS_UUID_SIZE];
+
 		item = btrfs_item_nr(i);
-		btrfs_item_key(l, &disk_key, i);
+		item_size = btrfs_item_size(eb, item);
+		/* Untyped extraction of slot from btrfs_item_ptr */
+		ptr = btrfs_item_ptr(eb, i, void*);
+
+		btrfs_item_key(eb, &disk_key, i);
 		objectid = btrfs_disk_key_objectid(&disk_key);
 		type = btrfs_disk_key_type(&disk_key);
+		offset = btrfs_disk_key_offset(&disk_key);
+
 		printf("\titem %d ", i);
 		btrfs_print_key(&disk_key);
 		printf(" itemoff %d itemsize %d\n",
-			btrfs_item_offset(l, item),
-			btrfs_item_size(l, item));
+			btrfs_item_offset(eb, item),
+			btrfs_item_size(eb, item));
 
 		if (type == 0 && objectid == BTRFS_FREE_SPACE_OBJECTID)
-			print_free_space_header(l, i);
+			print_free_space_header(eb, i);
 
 		switch (type) {
 		case BTRFS_INODE_ITEM_KEY:
-			memset(flags_str, 0, sizeof(flags_str));
-			ii = btrfs_item_ptr(l, i, struct btrfs_inode_item);
-			inode_flags_to_str(btrfs_inode_flags(l, ii), flags_str);
-			printf("\t\tinode generation %llu transid %llu size %llu nbytes %llu\n"
-			       "\t\tblock group %llu mode %o links %u uid %u gid %u\n"
-			       "\t\trdev %llu flags 0x%llx(%s)\n",
-			       (unsigned long long)btrfs_inode_generation(l, ii),
-			       (unsigned long long)btrfs_inode_transid(l, ii),
-			       (unsigned long long)btrfs_inode_size(l, ii),
-			       (unsigned long long)btrfs_inode_nbytes(l, ii),
-			       (unsigned long long)btrfs_inode_block_group(l,ii),
-			       btrfs_inode_mode(l, ii),
-			       btrfs_inode_nlink(l, ii),
-			       btrfs_inode_uid(l, ii),
-			       btrfs_inode_gid(l, ii),
-			       (unsigned long long)btrfs_inode_rdev(l,ii),
-			       (unsigned long long)btrfs_inode_flags(l,ii),
-			       flags_str);
+			print_inode_item(eb, ptr);
 			break;
 		case BTRFS_INODE_REF_KEY:
-			iref = btrfs_item_ptr(l, i, struct btrfs_inode_ref);
-			print_inode_ref_item(l, item, iref);
+			print_inode_ref_item(eb, item_size, ptr);
 			break;
 		case BTRFS_INODE_EXTREF_KEY:
-			iref2 = btrfs_item_ptr(l, i, struct btrfs_inode_extref);
-			print_inode_extref_item(l, item, iref2);
+			print_inode_extref_item(eb, item_size, ptr);
 			break;
 		case BTRFS_DIR_ITEM_KEY:
 		case BTRFS_DIR_INDEX_KEY:
 		case BTRFS_XATTR_ITEM_KEY:
-			di = btrfs_item_ptr(l, i, struct btrfs_dir_item);
-			print_dir_item(l, item, di);
+			print_dir_item(eb, item_size, ptr);
 			break;
 		case BTRFS_DIR_LOG_INDEX_KEY:
-		case BTRFS_DIR_LOG_ITEM_KEY:
-			dlog = btrfs_item_ptr(l, i, struct btrfs_dir_log_item);
+		case BTRFS_DIR_LOG_ITEM_KEY: {
+			struct btrfs_dir_log_item *dlog;
+
+			dlog = btrfs_item_ptr(eb, i, struct btrfs_dir_log_item);
 			printf("\t\tdir log end %Lu\n",
-			       (unsigned long long)btrfs_dir_log_end(l, dlog));
-		       break;
+			       (unsigned long long)btrfs_dir_log_end(eb, dlog));
+			break;
+			}
 		case BTRFS_ORPHAN_ITEM_KEY:
 			printf("\t\torphan item\n");
 			break;
 		case BTRFS_ROOT_ITEM_KEY:
-			print_root(l, i);
+			print_root(eb, i);
 			break;
 		case BTRFS_ROOT_REF_KEY:
-			print_root_ref(l, i, "ref");
+			print_root_ref(eb, i, "ref");
 			break;
 		case BTRFS_ROOT_BACKREF_KEY:
-			print_root_ref(l, i, "backref");
+			print_root_ref(eb, i, "backref");
 			break;
 		case BTRFS_EXTENT_ITEM_KEY:
-			print_extent_item(l, i, 0);
+			print_extent_item(eb, i, 0);
 			break;
 		case BTRFS_METADATA_ITEM_KEY:
-			print_extent_item(l, i, 1);
+			print_extent_item(eb, i, 1);
 			break;
 		case BTRFS_TREE_BLOCK_REF_KEY:
 			printf("\t\ttree block backref\n");
@@ -990,23 +1032,28 @@ void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
 		case BTRFS_SHARED_BLOCK_REF_KEY:
 			printf("\t\tshared block backref\n");
 			break;
-		case BTRFS_EXTENT_DATA_REF_KEY:
-			dref = btrfs_item_ptr(l, i, struct btrfs_extent_data_ref);
+		case BTRFS_EXTENT_DATA_REF_KEY: {
+			struct btrfs_extent_data_ref *dref;
+
+			dref = btrfs_item_ptr(eb, i, struct btrfs_extent_data_ref);
 			printf("\t\textent data backref root %llu "
 			       "objectid %llu offset %llu count %u\n",
-			       (unsigned long long)btrfs_extent_data_ref_root(l, dref),
-			       (unsigned long long)btrfs_extent_data_ref_objectid(l, dref),
-			       (unsigned long long)btrfs_extent_data_ref_offset(l, dref),
-			       btrfs_extent_data_ref_count(l, dref));
+			       (unsigned long long)btrfs_extent_data_ref_root(eb, dref),
+			       (unsigned long long)btrfs_extent_data_ref_objectid(eb, dref),
+			       (unsigned long long)btrfs_extent_data_ref_offset(eb, dref),
+			       btrfs_extent_data_ref_count(eb, dref));
 			break;
-		case BTRFS_SHARED_DATA_REF_KEY:
-			sref = btrfs_item_ptr(l, i, struct btrfs_shared_data_ref);
+			}
+		case BTRFS_SHARED_DATA_REF_KEY: {
+			struct btrfs_shared_data_ref *sref;
+			sref = btrfs_item_ptr(eb, i, struct btrfs_shared_data_ref);
 			printf("\t\tshared data backref count %u\n",
-			       btrfs_shared_data_ref_count(l, sref));
+			       btrfs_shared_data_ref_count(eb, sref));
 			break;
+			}
 		case BTRFS_EXTENT_REF_V0_KEY:
 #ifdef BTRFS_COMPAT_EXTENT_TREE_V0
-			print_extent_ref_v0(l, i);
+			print_extent_ref_v0(eb, i);
 #else
 			BUG();
 #endif
@@ -1018,14 +1065,12 @@ void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
 			printf("\t\textent csum item\n");
 			break;
 		case BTRFS_EXTENT_DATA_KEY:
-			fi = btrfs_item_ptr(l, i,
-					    struct btrfs_file_extent_item);
-			print_file_extent_item(l, item, i, fi);
+			print_file_extent_item(eb, item, i, ptr);
 			break;
-		case BTRFS_BLOCK_GROUP_ITEM_KEY:
-			bi = btrfs_item_ptr(l, i,
-					    struct btrfs_block_group_item);
-			read_extent_buffer(l, &bg_item, (unsigned long)bi,
+		case BTRFS_BLOCK_GROUP_ITEM_KEY: {
+			struct btrfs_block_group_item bg_item;
+
+			read_extent_buffer(eb, &bg_item, (unsigned long)ptr,
 					   sizeof(bg_item));
 			memset(flags_str, 0, sizeof(flags_str));
 			bg_flags_to_str(btrfs_block_group_flags(&bg_item),
@@ -1035,12 +1080,16 @@ void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
 			       (unsigned long long)btrfs_block_group_chunk_objectid(&bg_item),
 			       flags_str);
 			break;
-		case BTRFS_FREE_SPACE_INFO_KEY:
-			free_info = btrfs_item_ptr(l, i, struct btrfs_free_space_info);
+			}
+		case BTRFS_FREE_SPACE_INFO_KEY: {
+			struct btrfs_free_space_info *free_info;
+
+			free_info = btrfs_item_ptr(eb, i, struct btrfs_free_space_info);
 			printf("\t\tfree space info extent count %u flags %u\n",
-			       (unsigned)btrfs_free_space_extent_count(l, free_info),
-			       (unsigned)btrfs_free_space_flags(l, free_info));
+			       (unsigned)btrfs_free_space_extent_count(eb, free_info),
+			       (unsigned)btrfs_free_space_flags(eb, free_info));
 			break;
+			}
 		case BTRFS_FREE_SPACE_EXTENT_KEY:
 			printf("\t\tfree space extent\n");
 			break;
@@ -1048,93 +1097,134 @@ void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
 			printf("\t\tfree space bitmap\n");
 			break;
 		case BTRFS_CHUNK_ITEM_KEY:
-			print_chunk(l, btrfs_item_ptr(l, i, struct btrfs_chunk));
+			print_chunk(eb, ptr);
 			break;
 		case BTRFS_DEV_ITEM_KEY:
-			print_dev_item(l, btrfs_item_ptr(l, i,
-					struct btrfs_dev_item));
+			print_dev_item(eb, ptr);
 			break;
-		case BTRFS_DEV_EXTENT_KEY:
-			dev_extent = btrfs_item_ptr(l, i,
+		case BTRFS_DEV_EXTENT_KEY: {
+			struct btrfs_dev_extent *dev_extent;
+
+			dev_extent = btrfs_item_ptr(eb, i,
 						    struct btrfs_dev_extent);
+			read_extent_buffer(eb, uuid,
+				(unsigned long)btrfs_dev_extent_chunk_tree_uuid(dev_extent),
+				BTRFS_UUID_SIZE);
+			uuid_unparse(uuid, uuid_str);
 			printf("\t\tdev extent chunk_tree %llu\n"
-			       "\t\tchunk objectid %llu chunk offset %llu "
-			       "length %llu\n",
+			       "\t\tchunk_objectid %llu chunk_offset %llu "
+			       "length %llu\n"
+			       "\t\tchunk_tree_uuid %s\n",
 			       (unsigned long long)
-			       btrfs_dev_extent_chunk_tree(l, dev_extent),
+			       btrfs_dev_extent_chunk_tree(eb, dev_extent),
 			       (unsigned long long)
-			       btrfs_dev_extent_chunk_objectid(l, dev_extent),
+			       btrfs_dev_extent_chunk_objectid(eb, dev_extent),
 			       (unsigned long long)
-			       btrfs_dev_extent_chunk_offset(l, dev_extent),
+			       btrfs_dev_extent_chunk_offset(eb, dev_extent),
 			       (unsigned long long)
-			       btrfs_dev_extent_length(l, dev_extent));
+			       btrfs_dev_extent_length(eb, dev_extent),
+			       uuid_str);
 			break;
-		case BTRFS_QGROUP_STATUS_KEY:
-			qg_status = btrfs_item_ptr(l, i,
+			}
+		case BTRFS_QGROUP_STATUS_KEY: {
+			struct btrfs_qgroup_status_item *qg_status;
+
+			qg_status = btrfs_item_ptr(eb, i,
 					struct btrfs_qgroup_status_item);
 			memset(flags_str, 0, sizeof(flags_str));
-			qgroup_flags_to_str(btrfs_qgroup_status_flags(l, qg_status),
+			qgroup_flags_to_str(btrfs_qgroup_status_flags(eb, qg_status),
 					flags_str);
 			printf("\t\tversion %llu generation %llu flags %s "
 				"scan %lld\n",
 				(unsigned long long)
-				btrfs_qgroup_status_version(l, qg_status),
+				btrfs_qgroup_status_version(eb, qg_status),
 				(unsigned long long)
-				btrfs_qgroup_status_generation(l, qg_status),
+				btrfs_qgroup_status_generation(eb, qg_status),
 				flags_str,
 				(unsigned long long)
-				btrfs_qgroup_status_rescan(l, qg_status));
+				btrfs_qgroup_status_rescan(eb, qg_status));
 			break;
+			}
 		case BTRFS_QGROUP_RELATION_KEY:
 			break;
-		case BTRFS_QGROUP_INFO_KEY:
-			qg_info = btrfs_item_ptr(l, i,
+		case BTRFS_QGROUP_INFO_KEY: {
+			struct btrfs_qgroup_info_item *qg_info;
+
+			qg_info = btrfs_item_ptr(eb, i,
 						 struct btrfs_qgroup_info_item);
 			printf("\t\tgeneration %llu\n"
-			     "\t\treferenced %llu referenced compressed %llu\n"
-			     "\t\texclusive %llu exclusive compressed %llu\n",
+			     "\t\treferenced %llu referenced_compressed %llu\n"
+			     "\t\texclusive %llu exclusive_compressed %llu\n",
 			       (unsigned long long)
-			       btrfs_qgroup_info_generation(l, qg_info),
+			       btrfs_qgroup_info_generation(eb, qg_info),
 			       (unsigned long long)
-			       btrfs_qgroup_info_referenced(l, qg_info),
+			       btrfs_qgroup_info_referenced(eb, qg_info),
 			       (unsigned long long)
-			       btrfs_qgroup_info_referenced_compressed(l,
+			       btrfs_qgroup_info_referenced_compressed(eb,
 								       qg_info),
 			       (unsigned long long)
-			       btrfs_qgroup_info_exclusive(l, qg_info),
+			       btrfs_qgroup_info_exclusive(eb, qg_info),
 			       (unsigned long long)
-			       btrfs_qgroup_info_exclusive_compressed(l,
+			       btrfs_qgroup_info_exclusive_compressed(eb,
 								      qg_info));
 			break;
-		case BTRFS_QGROUP_LIMIT_KEY:
-			qg_limit = btrfs_item_ptr(l, i,
+			}
+		case BTRFS_QGROUP_LIMIT_KEY: {
+			struct btrfs_qgroup_limit_item *qg_limit;
+
+			qg_limit = btrfs_item_ptr(eb, i,
 					 struct btrfs_qgroup_limit_item);
 			printf("\t\tflags %llx\n"
-			     "\t\tmax referenced %lld max exclusive %lld\n"
-			     "\t\trsv referenced %lld rsv exclusive %lld\n",
+			     "\t\tmax_referenced %lld max_exclusive %lld\n"
+			     "\t\trsv_referenced %lld rsv_exclusive %lld\n",
 			       (unsigned long long)
-			       btrfs_qgroup_limit_flags(l, qg_limit),
+			       btrfs_qgroup_limit_flags(eb, qg_limit),
 			       (long long)
-			       btrfs_qgroup_limit_max_referenced(l, qg_limit),
+			       btrfs_qgroup_limit_max_referenced(eb, qg_limit),
 			       (long long)
-			       btrfs_qgroup_limit_max_exclusive(l, qg_limit),
+			       btrfs_qgroup_limit_max_exclusive(eb, qg_limit),
 			       (long long)
-			       btrfs_qgroup_limit_rsv_referenced(l, qg_limit),
+			       btrfs_qgroup_limit_rsv_referenced(eb, qg_limit),
 			       (long long)
-			       btrfs_qgroup_limit_rsv_exclusive(l, qg_limit));
+			       btrfs_qgroup_limit_rsv_exclusive(eb, qg_limit));
 			break;
+			}
 		case BTRFS_UUID_KEY_SUBVOL:
 		case BTRFS_UUID_KEY_RECEIVED_SUBVOL:
-			print_uuid_item(l, btrfs_item_ptr_offset(l, i),
-					btrfs_item_size_nr(l, i));
+			print_uuid_item(eb, btrfs_item_ptr_offset(eb, i),
+					btrfs_item_size_nr(eb, i));
 			break;
-		case BTRFS_STRING_ITEM_KEY:
-			/* dirty, but it's simple */
-			str = l->data + btrfs_item_ptr_offset(l, i);
-			printf("\t\titem data %.*s\n", btrfs_item_size(l, item), str);
+		case BTRFS_STRING_ITEM_KEY: {
+			const char *str = eb->data + btrfs_item_ptr_offset(eb, i);
+
+			printf("\t\titem data %.*s\n", item_size, str);
 			break;
-		case BTRFS_DEV_STATS_KEY:
-			printf("\t\tdevice stats\n");
+			}
+		case BTRFS_PERSISTENT_ITEM_KEY:
+			printf("\t\tpersistent item objectid ");
+			print_objectid(stdout, objectid, BTRFS_PERSISTENT_ITEM_KEY);
+			printf(" offset %llu\n", (unsigned long long)offset);
+			switch (objectid) {
+			case BTRFS_DEV_STATS_OBJECTID:
+				print_dev_stats(eb, ptr, item_size);
+				break;
+			default:
+				printf("\t\tunknown persistent item objectid %llu\n",
+						objectid);
+			}
+			break;
+		case BTRFS_TEMPORARY_ITEM_KEY:
+			printf("\t\ttemporary item objectid ");
+			print_objectid(stdout, objectid, BTRFS_TEMPORARY_ITEM_KEY);
+			printf(" offset %llu\n", (unsigned long long)offset);
+			switch (objectid) {
+			case BTRFS_BALANCE_OBJECTID:
+				print_balance_item(eb, ptr);
+				break;
+			default:
+				printf("\t\tunknown temporary item objectid %llu\n",
+						objectid);
+			}
 			break;
 		};
 		fflush(stdout);
@@ -1143,7 +1233,7 @@ void btrfs_print_leaf(struct btrfs_root *root, struct extent_buffer *l)
 
 void btrfs_print_tree(struct btrfs_root *root, struct extent_buffer *eb, int follow)
 {
-	int i;
+	u32 i;
 	u32 nr;
 	u32 size;
 	struct btrfs_disk_key disk_key;
