@@ -4,14 +4,14 @@
 # clean.
 
 LANG=C
-SCRIPT_DIR=$(dirname $(readlink -f $0))
-TOP=$(readlink -f $SCRIPT_DIR/../)
+SCRIPT_DIR=$(dirname $(readlink -f "$0"))
+TOP=$(readlink -f "$SCRIPT_DIR/../")
 TEST_DEV=${TEST_DEV:-}
 RESULTS="$TOP/tests/convert-tests-results.txt"
 IMAGE="$TOP/tests/test.img"
 
-source $TOP/tests/common
-source $TOP/tests/common.convert
+source "$TOP/tests/common"
+source "$TOP/tests/common.convert"
 
 export TOP
 export RESULTS
@@ -19,7 +19,9 @@ export LANG
 export IMAGE
 export TEST_DEV
 
-rm -f $RESULTS
+rm -f "$RESULTS"
+
+check_kernel_support
 
 run_one_test() {
 	local testdir
@@ -29,7 +31,7 @@ run_one_test() {
 	testname=$(basename "$testdir")
 	echo "    [TEST/conv]   $testname"
 	cd "$testdir"
-	echo "=== Entering $testname" >> $RESULTS
+	echo "=== Entering $testname" >> "$RESULTS"
 	if [ -x test.sh ]; then
 		# Only support custom test scripts
 		./test.sh
@@ -42,7 +44,7 @@ run_one_test() {
 }
 
 # Test special images
-for i in $(find $TOP/tests/convert-tests -maxdepth 1 -mindepth 1 -type d \
+for i in $(find "$TOP/tests/convert-tests" -maxdepth 1 -mindepth 1 -type d \
 	   ${TEST:+-name "$TEST"} | sort)
 do
 	run_one_test "$i"
