@@ -4,7 +4,6 @@
 source $TOP/tests/common
 
 check_prereq mkfs.btrfs
-check_prereq btrfs-show-super
 
 setup_root_helper
 prepare_test_dev
@@ -21,7 +20,8 @@ run_check truncate -s2g img
 loopdev=`run_check_stdout $SUDO_HELPER losetup --find --show img`
 run_check $SUDO_HELPER dmsetup create $dmname --table "0 1048576 linear $loopdev 0"
 
-base=`basename "$loopdev"`
+dmbase=`readlink -f $dmdev`
+base=`basename "$dmbase"`
 rot=/sys/class/block/$base/queue/rotational
 
 # switch rotational
@@ -32,7 +32,7 @@ run_check cat $rot
 # test
 run_check_stdout $SUDO_HELPER $TOP/mkfs.btrfs -f $@ $dmdev |
 	grep -q 'SSD detected:.*yes' || _fail 'SSD not detected'
-run_check $TOP/btrfs-show-super $dmdev
+run_check $TOP/btrfs inspect-internal dump-super $dmdev
 
 # cleanup
 run_check $SUDO_HELPER dmsetup remove $dmname
